(* Terminberechnungen

    2007 J. Rathlev 24147 Klausdorf
      Web:  www.rathlev-home.de 
      Mail: info(a)rathlev-home.de

   The contents of this file may be used under the terms of the
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Jan. 2005 *)

unit DateCalcDlg;

interface

uses Windows, SysUtils, Classes, Graphics, Forms, Controls, StdCtrls, 
  Buttons, ExtCtrls, ComCtrls;

type
  TDateCalcDialog = class(TForm)
    OKBtn: TBitBtn;
    CancelBtn: TBitBtn;
    Label1: TLabel;
    dtpStart: TDateTimePicker;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    edYear: TEdit;
    udYear: TUpDown;
    edMonth: TEdit;
    udMonth: TUpDown;
    edDay: TEdit;
    udDay: TUpDown;
    Label5: TLabel;
    dtpResult: TDateTimePicker;
    Label6: TLabel;
    edWeek: TEdit;
    udWeek: TUpDown;
    pcCalc: TPageControl;
    tsDate: TTabSheet;
    tsSpan: TTabSheet;
    dtpEnd: TDateTimePicker;
    Label7: TLabel;
    rgSpan: TRadioGroup;
    edSpan: TEdit;
    Label8: TLabel;
    btnToday: TBitBtn;
    procedure DateChange(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure btnTodayClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    function Execute (var ADate : TDateTime; var AHint : string) : boolean;
  end;

var
  DateCalcDialog: TDateCalcDialog;

implementation

{$R *.DFM}

uses DateUtils;

function TDateCalcDialog.Execute (var ADate : TDateTime; var AHint : string) : boolean;
begin
  dtpStart.Date:=ADate;
  dtpEnd.Date:=Date;
  if ShowModal=mrOK then begin
    if pcCalc.ActivePage=tsDate then begin
      ADate:=dtpResult.Date;
      AHint:=DateToStr(dtpStart.Date);
      with udYear do begin
        if Position>0 then AHint:=AHint+'+';
        if Position<>0 then AHint:=AHint+IntToStr(Position)+'y';
        end;
     with udMonth do begin
        if Position>0 then AHint:=AHint+'+';
        if Position<>0 then AHint:=AHint+IntToStr(Position)+'m';
        end;
     with udWeek do begin
        if Position>0 then AHint:=AHint+'+';
        if Position<>0 then AHint:=AHint+IntToStr(Position)+'w';
        end;
     with udDay do begin
        if Position>0 then AHint:=AHint+'+';
        if Position<>0 then AHint:=AHint+IntToStr(Position)+'d';
        end;
      end;
    Result:=true;
    end
  else Result:=false;
  end;


procedure TDateCalcDialog.DateChange(Sender: TObject);
var
  dt,ds,de  : TDateTime;
  s         : string;
  k,m,n     : integer;
  bf        : boolean;

  procedure GetDaysBetween(DStart,DEnd : TDateTime;
                           var before : boolean;
                           var days  : integer);
  begin
    before:=false;
    if SameDate(DStart,DEnd) then days:=0
    else begin
      if DStart>DEnd then begin
        before:=true;
        Days:=DaysBetween(DEnd,DStart+0.5);
        end
      else Days:=DaysBetween(DStart,DEnd);
      end;
    end;

  procedure GetWeeksBetween(DStart,DEnd : TDateTime;
                            var before : boolean;
                            var weeks,days  : integer);
  begin
    before:=false;
    if SameDate(DStart,DEnd) then days:=0
    else begin
      if DStart>DEnd then begin
        before:=true;
        Weeks:=WeeksBetween(DEnd,DStart+0.5);
        Days:=DaysBetween(IncWeek(DEnd,Weeks),DStart+0.5);
        end
      else begin
        Weeks:=WeeksBetween(DStart,DEnd);
        Days:=DaysBetween(IncWeek(DStart,Weeks),DEnd);
        end;
      end;
    end;

  procedure GetMonthsAndDaysBetween(DStart,DEnd : TDateTime;
                                    var before : boolean;
                                    var months,days  : integer);
  var
    dm : word;
  begin
    months:=0; before:=false;
    if SameDate(DStart,DEnd) then days:=0
    else begin
      if DStart>DEnd then begin
        before:=true;
        while DEnd<DStart+0.1 do begin
          dm:=DaysInMonth(DEnd);
          if DEnd+dm<DStart+0.1 then inc(Months)
          else Days:=DaysBetween(DEnd,DStart+0.1);
          DEnd:=DEnd+dm;
          end;
        end
      else begin
        while DStart<DEnd+0.1 do begin
          dm:=DaysInMonth(DStart);
          if DStart+dm<=DEnd then inc(Months)
          else Days:=DaysBetween(DStart,DEnd+0.1);
          DStart:=DStart+dm;
          end;
        end;
      end;
    end;

  procedure GetYearsAndDaysBetween(DStart,DEnd : TDateTime;
                                   var before : boolean;
                                   var years,months,days  : integer);
  var
    dm : word;
    b  : boolean;
  begin
    Years:=0; Months:=0; before:=false;
    if SameDate(DStart,DEnd) then days:=0
    else begin
      if DStart>DEnd then begin
        before:=true;
        while DEnd<DStart+0.5 do begin
          dm:=DaysInYear(DEnd);
          if DEnd+dm<DStart then inc(Years)
          else GetMonthsAndDaysBetween(DEnd,DStart+0.5,b,Months,Days);
          DEnd:=DEnd+dm;
          end;
        end
      else begin
        while DStart<DEnd do begin
          dm:=DaysInYear(DStart);
          if DStart+dm<=DEnd then inc(Years)
          else GetMonthsAndDaysBetween(DStart,DEnd,b,Months,Days);
          DStart:=DStart+dm;
          end;
        end;
      end;
    end;

begin
  if pcCalc.ActivePage=tsDate then begin
    dt:=IncYear(dtpStart.Date,udYear.Position);
    dt:=IncMonth(dt,udMonth.Position);
    dt:=IncWeek(dt,udWeek.Position);
    dtpResult.Date:=IncDay(dt,udDay.Position);
    okBtn.Visible:=true;
    end
  else begin
    ds:=DateOf(dtpStart.Date);
    de:=DateOf(dtpEnd.Date);
    case rgSpan.ItemIndex of
    1 : begin
        GetWeeksBetween(ds,de,bf,m,n);
        s:=IntToStr(m)+'w '+IntToStr(n)+'d';
        end;
    2 : begin
        GetMonthsAndDaysBetween(ds,de,bf,m,n);
        s:=IntToStr(m)+'m '+IntToStr(n)+'d';
        end;
    3 : begin
        GetYearsAndDaysBetween(ds,de,bf,k,m,n);
        s:=IntToStr(k)+'y '+IntToStr(m)+'m '+IntToStr(n)+'d';
        end;
    else begin
        GetDaysBetween(ds,de,bf,n);
        s:=IntToStr(n)+'d';
        end;
      end;
    if bf then s:='- '+s;  
    edSpan.Text:=s;
    okBtn.Visible:=false;
    end
  end;

procedure TDateCalcDialog.FormActivate(Sender: TObject);
begin
  DateChange(Sender);
  end;

procedure TDateCalcDialog.btnTodayClick(Sender: TObject);
begin
  dtpEnd.Date:=Date;
  DateChange(Sender);
  end;

end.
