(* Delphi Dialog
   Eingabe und Auswahl von Feiertagen fr einen Kalender
   =====================================================

    Dr. J. Rathlev, D-24222 Schwentinental (kontakt(a)rathlev-home.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - Jan. 2005
   Vers. 2 - Sep. 2010
   letzte nderung:  Jan. 2024
   *)

unit EditHolidayDlg;

interface

uses Winapi.Windows, System.SysUtils, System.Classes, Vcl.Graphics, Vcl.Forms,
  Vcl.Controls, Vcl.StdCtrls, Vcl.Buttons, Vcl.ExtCtrls, Vcl.ComCtrls,
  NumberEd, DateUtilEx;

type
  THolidayObject = class (TObject)
  private
    FData   : THoliday;
    procedure SetHolidayName (const Value : string);
    procedure SetHolidayType (Value : THolidayType);
  public
    constructor Create (const AData : THoliday);
    property Data : THoliday read FData write FData;
    property HolidayName : string write SetHolidayName;
    property HolidayType : THolidayType write SetHolidayType;
    end;

  TEditHolidayDialog = class(TForm)
    OKBtn: TBitBtn;
    CancelBtn: TBitBtn;
    cbxHolidays: TComboBox;
    rgpType: TRadioGroup;
    pctType: TPageControl;
    btnNew: TBitBtn;
    btnDelete: TBitBtn;
    tshDate: TTabSheet;
    tshOffset: TTabSheet;
    Label1: TLabel;
    edtDay: TRangeEdit;
    Label3: TLabel;
    tshMonth: TTabSheet;
    Label5: TLabel;
    cbMWeekday: TComboBox;
    rbFirst: TRadioButton;
    rbLast: TRadioButton;
    btnEdit: TBitBtn;
    cbMonth1: TComboBox;
    Label7: TLabel;
    cbMonth2: TComboBox;
    rbSecond: TRadioButton;
    rbNextLast: TRadioButton;
    Label8: TLabel;
    tshWeek: TTabSheet;
    cbWeekday: TComboBox;
    Label2: TLabel;
    Label4: TLabel;
    Label6: TLabel;
    Label9: TLabel;
    rgbView: TRadioGroup;
    Panel2: TPanel;
    reMult: TRangeEdit;
    udMult: TNumUpDown;
    reOffs: TRangeEdit;
    udOffs: TNumUpDown;
    udWeekDist: TNumUpDown;
    reWeekDist: TRangeEdit;
    reFirstWeek: TRangeEdit;
    udFirstWeek: TNumUpDown;
    btnNext: TBitBtn;
    btnPrevious: TBitBtn;
    procedure rgpTypeClick(Sender: TObject);
    procedure cbxHolidaysChange(Sender: TObject);
    procedure btnNewClick(Sender: TObject);
    procedure btnDeleteClick(Sender: TObject);
    procedure btnEditClick(Sender: TObject);
    procedure btnNextClick(Sender: TObject);
    procedure btnPreviousClick(Sender: TObject);
  private
    { Private declarations }
    LastIndex : integer;
{$IFDEF HDPI}   // scale glyphs and images for High DPI
    procedure AfterConstruction; override;
{$EndIf}
    procedure UpdateFocus;
    procedure ShowDate (const h : THoliday);
    procedure ShowHoliDay (AIndex : integer);
  public
    { Public declarations }
    function Execute (var AHolidayCalendar : THolidayCalendar; var AName : string) : boolean;
  end;

var
  EditHolidayDialog: TEditHolidayDialog;

implementation

{$R *.DFM}

uses Vcl.Dialogs, WinUtils, ListUtils, MsgDialogs;

{------------------------------------------------------------------- }
constructor THolidayObject.Create (const AData : THoliday);
begin
  inherited Create;
  FData:=AData;
  end;

procedure THolidayObject.SetHolidayName (const Value : string);
begin
  FData.Name:=Value;
  end;

procedure THolidayObject.SetHolidayType (Value : THolidayType);
begin
  FData.HType:=Value;
  end;

{------------------------------------------------------------------- }
{$IFDEF HDPI}   // scale glyphs and images for High DPI
procedure THolidayObject.AfterConstruction;
begin
  inherited;
  if Application.Tag=0 then
    ScaleButtonGlyphs(self,PixelsPerInchOnDesign,Monitor.PixelsPerInch);
  end;
{$EndIf}

{------------------------------------------------------------------- }
procedure TEditHolidayDialog.cbxHolidaysChange(Sender : TObject);
begin
  ShowHoliDay(cbxHolidays.ItemIndex);
  end;

procedure TEditHolidayDialog.ShowHoliDay (AIndex : integer);
var
  h : THoliday;
begin
  if (cbxHolidays.Items.Count>0) and (LastIndex>=0) then begin
    h:=(cbxHolidays.Items.Objects[LastIndex] as THolidayObject).Data;
    with h do begin
      HType:=THolidayType(rgpType.ItemIndex);
      HView:=THolidayView(rgbView.ItemIndex+1);
      if HType=htFix then begin
        Day:=(cbMonth1.ItemIndex+1) shl 8 or edtDay.Value;
        end
      else if HType=htMonth then begin
        Day:=(cbMonth2.ItemIndex+1) shl 8 or cbMWeekday.ItemIndex; // Mo = 0
        if rbLast.Checked then Day:=Day or $10
        else if rbSecond.Checked then Day:=Day or $20
        else if rbNextLast.Checked then Day:=Day or $30;
        end
      else if HType=htWeek then begin
        Day:=reFirstWeek.Value shl 16 or reWeekDist.Value shl 8 or cbWeekDay.ItemIndex;
        end
      else begin
        Offset:=reOffs.Value;
        end;
      Mult:=reMult.Value;
      end;
    (cbxHolidays.Items.Objects[LastIndex] as THolidayObject).Data:=h;
    end;
  if AIndex>=0 then with cbxHolidays do begin
    LastIndex:=AIndex;
    ItemIndex:=AIndex;
    h:=(Items.Objects[ItemIndex] as THolidayObject).Data;
    rgpType.ItemIndex:=integer(h.HType);
    rgbView.ItemIndex:=integer(h.HView)-1;
    Showdate(h);
    end;
  UpdateFocus;
  end;

procedure TEditHolidayDialog.rgpTypeClick(Sender: TObject);
var
  h : THoliday;
begin
  with cbxHolidays,(cbxHolidays.Items.Objects[LastIndex] as THolidayObject) do begin
    HolidayType:=THolidayType(rgpType.ItemIndex);
    ShowDate(Data);
    end;
  end;

procedure TEditHolidayDialog.ShowDate (const h : THoliday);
begin
  with reMult do begin
    MinValue:=-99; MaxValue:=99;
    end;
  if rgpType.ItemIndex=0 then begin     // festes Datum
    pctType.ActivePageIndex:=0;
    with h do begin
      edtDay.Value:=Day and $FF;
      cbMonth1.Itemindex:=(Day shr 8)and $FF-1;
      reMult.Value:=Mult;
      end;
    if Active then edtDay.SetFocus;
    end
  else if rgpType.ItemIndex=3 then begin   // Wochentag im Monat
    pctType.ActivePageIndex:=2;
    with h do begin
      cbMonth2.ItemIndex:=(Day shr 8)and $FF-1;
      cbMWeekday.ItemIndex:=Day and $F;
      case Day shr 4 and $3 of
      1 : rbLast.Checked:=true;
      2 : rbSecond.Checked:=true;
      3 : rbNextLast.Checked:=true;
        else rbFirst.Checked:=true
        end;
      reMult.Value:=Mult;
      end;
    end
  else if rgpType.ItemIndex=4 then begin   // wiederkehrender Wochentag
    pctType.ActivePageIndex:=3;
    with h do begin
      cbWeekDay.ItemIndex:=Day and $F;
      reWeekDist.Value:=Day shr 8 and $3F;
      reFirstWeek.Value:=Day shr 16 and $3F;
      with reMult do begin
        MinValue:=0; MaxValue:=5;
        Value:=Mult;
        end;
      end;
    end
  else begin                          // Ostern oder Advent
    pctType.ActivePageIndex:=1;
    with h do begin
      reOffs.Value:=Offset;
      reMult.Value:=Mult;
      end;
    if Active then reOffs.SetFocus;
    end;
  end;

procedure TEditHolidayDialog.UpdateFocus;
begin
  if Visible then begin
    if pctType.ActivePage=tshDate then edtDay.SetFocus
    else if pctType.ActivePage=tshOffset then reOffs.SetFocus
    else if pctType.ActivePage=tshWeek then reWeekDist.SetFocus
    else cbMonth2.SetFocus;
    end;
  end;

{------------------------------------------------------------------- }
procedure TEditHolidayDialog.btnNewClick(Sender: TObject);
var
  s : string;
  h : THoliday;
begin
  if InputQuery ('Neuer Feiertag oder Termin','Name:',s) then begin
    with h do begin
      Name:=s; HView:=hvNone; HType:=htFix; Day:=$101; Mult:=0; Offset:=0;
      Active:=true;
      end;
    ShowHoliDay(cbxHolidays.Items.AddObject(s,THolidayObject.Create(h)));
    end;
  end;

procedure TEditHolidayDialog.btnEditClick(Sender: TObject);
var
  sn : string;
  ho : THolidayObject;
begin
  with cbxHolidays do if (Items.Count>0) then begin
    ho:=(Items.Objects[ItemIndex] as THolidayObject);
    sn:=ho.Data.Name;
    if InputQuery ('Feiertag oder Termin umbenennen','Name:',sn) then begin
      ho.HolidayName:=sn;
      with Items do begin
        Delete(ItemIndex);
        ShowHoliDay(AddObject(sn,ho));
        end;
      end;
    end;
  end;

procedure TEditHolidayDialog.btnDeleteClick(Sender: TObject);
var
  n : integer;
begin
  with cbxHolidays do if (Items.Count>0) and
    ConfirmDialog('',Format('"%s"'+sLineBreak+'aus der Liste entfernen?',[Items[ItemIndex]])) then begin
    n:=ItemIndex;
    with Items do begin
      Objects[n].Free; Delete(n);
      LastIndex:=-1;
      end;
    if Items.Count<=0 then begin
      n:=-1; Text:='';
      end
    else if n>=Items.Count then n:=Items.Count-1;
    ShowHoliDay(n);
    end;
  end;

procedure TEditHolidayDialog.btnNextClick(Sender: TObject);
begin
  with cbxHolidays do if ItemIndex<Items.Count-1 then ShowHoliDay(ItemIndex+1);
  end;

procedure TEditHolidayDialog.btnPreviousClick(Sender: TObject);
begin
  with cbxHolidays do if ItemIndex>0 then ShowHoliDay(ItemIndex-1);
  end;

{------------------------------------------------------------------- }
(* Dialog an Position anzeigen *)
function TEditHolidayDialog.Execute (var AHolidayCalendar : THolidayCalendar;
                                     var AName : string) : boolean;
var
  i,n   : integer;
  h     : THolidayObject;
  AutoList : TStringList;
begin
  LastIndex:=-1;
  cbxHolidays.Clear; AutoList:=TStringList.Create;;
  with AHolidayCalendar do for i:=0 to HolidayCount-1 do begin
    h:=THolidayObject.Create (Holiday[i]);
    if Holiday[i].HType=htAuto then AutoList.AddObject(Holiday[i].Name,h)
    else cbxHolidays.Items.AddObject(Holiday[i].Name,h);
    end;
  with cbxHolidays do begin
    n:=Items.IndexOf(AName);
    if n>=0 then ItemIndex:=n
    else if Items.Count>0 then ItemIndex:=0 else ItemIndex:=-1;
    ShowHoliDay(ItemIndex);
    end;
  if ShowModal=mrOK then begin
    ShowHoliDay(-1);  // letzte nderung speichern
    with cbxHolidays do AName:=Items[ItemIndex];
    with AHolidayCalendar do begin
      HolidayCount:=AutoList.Count+cbxHolidays.Items.Count; n:=0;
      with cbxHolidays do for i:=0 to Items.Count-1 do with (Items.Objects[i] as THolidayObject) do begin
        Holiday[n]:=Data; inc(n);
        end;
      with AutoList do for i:=0 to Count-1 do with (Objects[i] as THolidayObject) do begin
        Holiday[n]:=Data; inc(n);
        end;
      Update;
      end;
    Result:=true;
    end
  else Result:=false;
  FreeListObjects(cbxHolidays.Items);
  FreeListObjects(AutoList); AutoList.Free;
  end;

end.
