(* Kalender mit Feiertagen

    2016-2024 J. Rathlev D-24222 Schwentinental
      Web:  www.rathlev-home.de
      Mail: kontakt(a)rathlev-home.de

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Jan. 2005, Aug. 2008
   letzte nderung:  Jan. 2024
   *)

unit KalMain;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes,
  Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Menus, Vcl.ExtCtrls, System.DateUtils,
  Vcl.Dialogs, Vcl.Grids, Vcl.Samples.Calendar, Vcl.StdCtrls, Vcl.ComCtrls, Vcl.Buttons,
  System.IniFiles, JRCalendar, DateUtilEx, WinApiUtils, CClock, NumberEd,
  KalSettings, System.ImageList, Vcl.ImgList;


const
  Prog = 'Kalender mit Feiertagen';
  Vers = ' - Vers. 4.5';
  sAuto = 'auto';
  sIni  = 'ini';

  UserId = 1;
  WM_TASKBAREVENT = WM_USER+1;
  WM_SHOW = WM_USER+1987;

{$I JrGlobal.pas}
{ enthlt:
  CopRgt = ' 2014 - 2024 J. Rathlev, D-24222 Schwentinental';
  EMailAdr = 'kontakt@rathlev-home.de';
  }

type
  TfrmKalender = class(TForm)
    JRCalendar1: TJRCalendar;
    JRCalendar2: TJRCalendar;
    JRCalendar3: TJRCalendar;
    JRCalendar4: TJRCalendar;
    JRCalendar5: TJRCalendar;
    JRCalendar6: TJRCalendar;
    panMonth1: TPanel;
    panMonth2: TPanel;
    panMonth3: TPanel;
    panMonth4: TPanel;
    panMonth5: TPanel;
    panMonth6: TPanel;
    btnQMinus: TBitBtn;
    btnYPlus: TBitBtn;
    btnQPlus: TBitBtn;
    btnYPMinus: TBitBtn;
    lbxHolidays: TListBox;
    btnToday: TBitBtn;
    PopupMenu: TPopupMenu;
    KalenderItem: TMenuItem;
    N1: TMenuItem;
    CloseItem: TMenuItem;
    Abbrechen1: TMenuItem;
    paClock: TPanel;
    Clock: TClock;
    btnTime: TBitBtn;
    btnCalc: TBitBtn;
    udJahr: TNumUpDown;
    edJahr: TNumberEdit;
    btnPrint: TBitBtn;
    paSunMoon: TPanel;
    paMain: TPanel;
    Label1: TLabel;
    Label2: TLabel;
    edSun: TEdit;
    gbDay: TGroupBox;
    edMoon: TEdit;
    ilMoonPhases: TImageList;
    ilStarSigns: TImageList;
    imgMoonPhase: TImage;
    imgSunSign: TImage;
    Label3: TLabel;
    btnInfo: TBitBtn;
    btnExit: TBitBtn;
    bbSettings: TBitBtn;
    procedure FormShow(Sender: TObject);
    procedure btnQPlusClick(Sender: TObject);
    procedure btnQMinusClick(Sender: TObject);
    procedure btnYPMinusClick(Sender: TObject);
    procedure btnYPlusClick(Sender: TObject);
    procedure btnEndeClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure btnSettingsClick(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btnInfoClick(Sender: TObject);
    procedure JRCalendarSelectCell(Sender: TObject; ACol, ARow: Integer;
      var CanSelect: Boolean);
    procedure btnTodayClick(Sender: TObject);
    procedure btnYearClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure CloseItemClick(Sender: TObject);
    procedure KalenderItemClick(Sender: TObject);
    procedure btnTimeClick(Sender: TObject);
    procedure btnCalcClick(Sender: TObject);
    procedure edJahrKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure edJahrKeyPress(Sender: TObject; var Key: Char);
    procedure JRCalendarMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure btnPrintClick(Sender: TObject);
    procedure udJahrClick(Sender: TObject);
    procedure edJahrExit(Sender: TObject);
    procedure FormMouseWheelDown(Sender: TObject; Shift: TShiftState;
      MousePos: TPoint; var Handled: Boolean);
    procedure FormMouseWheelUp(Sender: TObject; Shift: TShiftState;
      MousePos: TPoint; var Handled: Boolean);
    procedure lbxHolidaysMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
  private
    { Private-Deklarationen }
    IniName,
    PrgName,
    ProgVersName,
    ProgVersDate,
    DataPath        : string;
    LastHolIndex,
    DefHgt          : integer;
    AJahr,AMonat    : word;
    Options         : TKalOptions;
    VersInfo        : TFileVersionInfo;
    HolidayCalendar : THolidayCalendar;
    InAuto,
    InSystray       : boolean;
    IconHandle      : HICON;
    ShowDate        : TDateTime;
    MarkHint        : string;
    HintWin         : THintWindow;
    procedure SaveToIni;
    procedure ShowHolidays (Jahr : word);
    procedure ShowKal (Monat,Jahr : word); overload;
    procedure ShowKal (ADate : TDateTime); overload;
    procedure ShowSunMoon;
    procedure UpdateHolidays;
    procedure SetMarkDate (ADate : TDateTime);
    procedure UpdateAutoStart;
    procedure WMTASKBAREVENT (var message: TMessage); message WM_TASKBAREVENT;
    procedure MsgHandler (var Msg : TMsg; var Handled : boolean);
  public
    { Public-Deklarationen }
    procedure ShowForm;
  end;

var
  frmKalender: TfrmKalender;

implementation

uses Winapi.ShlObj, System.StrUtils, System.Math, WinUtils, MsgDialogs, WinShell, SunMoon,
  InitProg, NumberUtils, StringUtils, InpNumber, PathUtils,
  TimeZoneDlg, DateCalcDlg, PrintPreviewDlg, LocalityDlg, EditHolidayDlg;

{$R *.dfm}
{$R KalIcon.res}

const
  IniExt = 'ini';

  (* INI-Sektionen *)
  CfGSekt = 'Config';

  iniTop  = 'Top';
  iniLeft = 'Left';
  iniOptions = 'Options';
  iniPath = 'DataPath';
  iniLoc  = 'Location';
  iniLat  = 'Latitidue';
  iniLon  = 'Longitude';
  iniTz   = 'TimeZone';
  iniDs   = 'DaylightSaving';

  StarSigns : array [TStarSign] of string = (
    'Widder','Stier','Zwillinge','Krebs','Lwe','Jungfrau','Waage','Skorpion',
    'Schtze','Steinbock','Wassermann','Fische');

  MoonPhases : array [TLunarPhase] of string = (
    'Neumond', 'Zunehmende Sichel', 'Zunehmender Halbmond', 'Zunehmender Mond',
    'Vollmond', 'Abnehmender Mond', 'Abnehmender Halbmond', 'Abnehmende Sichel');

procedure TfrmKalender.FormCreate(Sender: TObject);
var
  IniFile  : TMemIniFile;
  w        : word;
  i,sz     : integer;
  fn,s     : string;
  PosInfo  : TPosInfo;
begin
  with PosInfo do begin
    TimeZone:=1; UseDaylightSaving:=false;
    Latitude:=50; Longitude:=15*TimeZone;
    end;
  PrgName:=ExtractFileName(DelExt(Application.ExeName));
  IniName:='';
  InAuto:=false; // (ParamCount>0) and (AnsiLowerCase(ParamStr(1))=Auto);
  if ParamCount>0 then begin
    for i:=1 to ParamCount do begin
      s:=ParamStr(i);
      if (s[1]='/') or (s[1]='-') then begin
        delete (s,1,1);
        if CompareOption(s,sAuto) then InAuto:=true
        else if ReadOptionValue(s,sIni) then   // anderer Ort fr Ini-Datei
          IniName:=AddExt(ExpandRelativePath(ExtractFilePath(Application.ExeName),s),IniExt);
        end
      end;
    end;
  if length(IniName)=0 then IniName:=Erweiter(GetDesktopFolder(CSIDL_APPDATA),PrgName,IniExt);
  IniFile:=TMemIniFile.Create(IniName);
  with IniFile do begin
    Top:=ReadInteger (CfgSekt,iniTop,Top);
    Left:=ReadInteger (CfgSekt,iniLeft,Left);
    WordToSet(ReadInteger (CfgSekt,iniOptions,0),@Options);
    DataPath:=ReadString(CfgSekt,iniPath,ExtractFilePath(Application.ExeName));
    with PosInfo do begin
      LocName:=ReadString(CfgSekt,iniLoc,'');
      Latitude:=ReadFloat(CfgSekt,iniLat,Latitude);
      Longitude:=ReadFloat(CfgSekt,iniLon,Longitude);
      TimeZone:=ReadInteger(CfgSekt,iniTz,TimeZone);
      UseDaylightSaving:=ReadBool(CfgSekt,iniDs,UseDaylightSaving);
      end;
    Free;
    end;
//  if length(DataPath)=0 then DataPath:=GetDesktopFolder(CSIDL_PERSONAL);
  with VersInfo do if GetFileVersion (Application.ExeName,VersInfo) then begin
    // Versionsnr. ohne "Compilierung"
    Comments:='Vers. '+DelExt(Version);
    end
  else begin
    InternalName:=Prog; Version:=Vers; Copyright:=CopRgt;
    end;
  with VersInfo do ProgVersName:=InternalName+' ('+Comments+')';
  ProgVersDate:=DateToStr(FileDateToDateTime(FileAge(Application.ExeName)));
  IconHandle:=LoadIcon(hInstance,'SMALLICON');
  HintWin:=THintWindow.Create(self);
  DefHgt:=ClientHeight;
  DecodeDate(Date,AJahr,AMonat,w);
  AMonat:=3*((AMonat-1) div 3)+1;
  HolidayCalendar:=THolidayCalendar.Create(PosInfo,AJahr);
  JRCalendar1.HolidayView:=HolidayCalendar.GetHolidayView;
  JRCalendar2.HolidayView:=HolidayCalendar.GetHolidayView;
  JRCalendar3.HolidayView:=HolidayCalendar.GetHolidayView;
  JRCalendar4.HolidayView:=HolidayCalendar.GetHolidayView;
  JRCalendar5.HolidayView:=HolidayCalendar.GetHolidayView;
  JRCalendar6.HolidayView:=HolidayCalendar.GetHolidayView;
  if IsVista then begin
    fn:='Segoe UI'; sz:=9;
    end
  else begin
    fn:='Microsoft Sans Serif'; sz:=8;
    end;
  ShowDate:=Date; MarkHint:='';
  for i:=0 to ComponentCount-1 do begin
    if Components[i] is TPanel then (Components[i] as TPanel).ParentBackground:=false;
    if Components[i] is TJRCalendar then with (Components[i] as TJRCalendar).Font do begin
      Name:=fn; Size:=sz;
      end;
    end;
  with lbxHolidays do begin
    with Font do begin
      Name:=fn; Size:=sz;
      end;
    Tag:=0;
    end;
  LastHolIndex:=-1;
// Message-Handler fr Application - siehe Mutex
  with Application do begin
    OnMessage:=MsgHandler;
    end;
  end;

procedure TfrmKalender.ShowForm;
begin
  if not (koSunMoon in Options) then begin
    ClientHeight:=DefHgt-paSunMoon.Height;
    paSunMoon.Visible:=false;
    end;
  HolidayCalendar.InitData(IniName,koSeasons in Options,koMoonCycles in Options,hvHoliday);
  if InAuto and (koSysTray in Options) then begin
    TaskBarAddIcon (Handle,UserId,WM_TASKBAREVENT,IconHandle,'','',ProgVersName);
    InSystray:=true;
    Application.ShowMainForm:=false;
    Visible:=false;
    end
  else begin
    InSystray:=false;
    Application.ShowMainForm:=true;
    Visible:=true;
    end;
  end;

procedure TfrmKalender.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  SaveToIni;
  HolidayCalendar.SaveData;
  TaskBarRemoveIcon (handle,UserId);
  UpdateAutoStart;
  end;

procedure TfrmKalender.FormDestroy(Sender: TObject);
begin
  HolidayCalendar.Free;
  HintWin.Free;
  end;

procedure TfrmKalender.SaveToIni;
var
  IniFile  : TMemIniFile;
begin
  IniFile:=TMemIniFile.Create(IniName);
  with IniFile do begin
    WriteInteger (CfgSekt,iniTop,Top);
    WriteInteger (CfgSekt,iniLeft,Left);
    WriteInteger (CfgSekt,iniOptions,SetToWord(@Options));
    WriteString (CfgSekt,iniPath,DataPath);
    with HolidayCalendar do begin
      with PosInfo do begin
        WriteString(CfgSekt,iniLoc,LocName);
        WriteFloat(CfgSekt,iniLat,Latitude);
        WriteFloat(CfgSekt,iniLon,Longitude);
        WriteInteger(CfgSekt,iniTz,TimeZone);
        WriteBool(CfgSekt,iniDs,UseDaylightSaving);
        end;
      end;
    UpdateFile;
    Free;
    end;
//  HolidayCalendar.SaveToIni;
  end;

procedure TfrmKalender.ShowHolidays (Jahr : word);
var
  s       : string;
  ndx,j,m,
  i,k,n   : integer;
  fnd     : boolean;
  dt      : TDateTime;
  hfnd    : array of integer;

  function NotFound (j,k : integer) : boolean;
  var
    i : integer;
  begin
    Result:=false;
    for i:=0 to j-1 do if hfnd[i]=k then Exit;
    Result:=true;
    end;

begin
  with lbxHolidays do if Tag<>Jahr then begin
    Tag:=Jahr;
    ndx:=ItemIndex;
    Clear;
    with HolidayCalendar do begin
      SetLength(hfnd,HolidayCount);
      ChangeYear(Jahr);
      dt:=EncodeDate(Jahr,1,1);
      i:=0; j:=0; k:=-1; n:=DaysInYear(dt);
      repeat
        repeat
          k:=FindNextMatchingHoliday(dt+i,k+1);
          fnd:=(k>=0);
          if fnd then begin
            m:=GetHolidayIndex(k);
            fnd:=Holiday[m].Active and (Holiday[m].HView in hvShow);
            end;
          until (k<0) or fnd;
        if fnd and (((Holiday[m].Mult=0) and (Holiday[m].HType<>htWeek)) or NotFound(j,m)) then with Holiday[m] do begin
          hfnd[j]:=m; inc(j);
          if Mult=0 then s:=DayMonthToStr(dt+i)
          else begin
            s:=DayMonthToStr(dt+i);
            if Mult<0 then begin
              if Mult<-1 then s:=s+'-'+DayMonthToStr(dt+i-Mult)
              else  s:=s+'/'+DayMonthToStr(dt+i-Mult);
              end
            else begin
              if Mult>1 then s:=s+'-'+DayMonthToStr(dt+i+Mult)
              else s:=s+'/'+DayMonthToStr(dt+i+1);
//              i:=i+Mult;
              end;
            end;
          if HType=htWeek then s:=s+' ('+IntToStr(Day shr 8 and $3f)+' W)';
          Items.AddObject(Name+':  '+s,pointer(round(DateOf(dt+i))));
          end
        else inc(i);
        until i>=n;
      hfnd:=nil;
      end;
    ItemIndex:=ndx;  
    end;
  end;

procedure TfrmKalender.SetMarkDate (ADate : TDateTime);
begin
  JRCalendar1.MarkDate:=ADate;
  JRCalendar2.MarkDate:=ADate;
  JRCalendar3.MarkDate:=ADate;
  JRCalendar4.MarkDate:=ADate;
  JRCalendar5.MarkDate:=ADate;
  JRCalendar6.MarkDate:=ADate;
//  ShowKal(ADate);
  end;

procedure TfrmKalender.ShowKal (Monat,Jahr : word);
begin
  ShowHolidays (Jahr);
  edJahr.Value:=Jahr;
  with JRCalendar1 do begin
    Year:=Jahr; Month:=Monat;
    end;
  panMonth1.Caption:=FormatSettings.LongMonthNames[Monat];
  inc(Monat);
  with JRCalendar2 do begin
    Year:=Jahr; Month:=Monat;
    end;
  panMonth2.Caption:=FormatSettings.LongMonthNames[Monat];
  inc(Monat);
  with JRCalendar3 do begin
    Year:=Jahr; Month:=Monat;
    end;
  panMonth3.Caption:=FormatSettings.LongMonthNames[Monat];
  inc(Monat);
  if Monat>12 then begin
    inc(Jahr); Monat:=1;
    end;
  with JRCalendar4 do begin
    Year:=Jahr; Month:=Monat;
    end;
  panMonth4.Caption:=FormatSettings.LongMonthNames[Monat];
  inc(Monat);
  with JRCalendar5 do begin
    Year:=Jahr; Month:=Monat;
    end;
  panMonth5.Caption:=FormatSettings.LongMonthNames[Monat];
  inc(Monat);
  with JRCalendar6 do begin
    Year:=Jahr; Month:=Monat;
    end;
  panMonth6.Caption:=FormatSettings.LongMonthNames[Monat];
  UpdateHolidays;
  ShowSunMoon;
  end;

procedure TfrmKalender.ShowKal (ADate : TDateTime);
var
  w : word;
begin
  DecodeDate(ADate,AJahr,AMonat,w);
  AMonat:=3*((AMonat-1) div 3)+1;
  ShowKal (AMonat,AJahr);
  end;

procedure TfrmKalender.FormShow(Sender: TObject);
begin
  PrintPreviewDialog.LoadFromIni(IniName);
  with HintWin do begin
    Canvas.Font:=JrCalendar1.Font;
    Brush.Color:=clInfoBk;
    end;
  ShowKal (AMonat,AJahr);
//  Application.BringToFront;
  end;

procedure TfrmKalender.btnQPlusClick(Sender: TObject);
begin
  if AMonat<10 then inc(AMonat,3)
  else begin
    inc(AJahr); AMonat:=1;
    end;
  ShowKal (AMonat,AJahr);
  end;

procedure TfrmKalender.btnQMinusClick(Sender: TObject);
begin
  if AMonat>3 then dec(AMonat,3)
  else begin
    dec(AJahr); AMonat:=10;
    end;
  ShowKal (AMonat,AJahr);
  end;

procedure TfrmKalender.btnYPlusClick(Sender: TObject);
begin
  inc(AJahr);
  ShowKal (AMonat,AJahr);
  end;

procedure TfrmKalender.btnYPMinusClick(Sender: TObject);
begin
  dec(AJahr);
  ShowKal (AMonat,AJahr);
  end;

procedure TfrmKalender.btnEndeClick(Sender: TObject);
begin
  if (koSysTray in Options) then begin
    SaveToIni;
    if not InSystray then TaskBarAddIcon (Handle,UserId,WM_TASKBAREVENT,IconHandle,'','',ProgVersName);
    InSystray:=true;
    Hide;
    end
  else Close;
  end;

procedure TfrmKalender.btnSettingsClick(Sender: TObject);
begin
  if KalSettingsDialog.Execute(DataPath,HolidayCalendar,Options) then begin
    paSunMoon.Visible:=koSunMoon in Options;
    if paSunMoon.Visible then ClientHeight:=DefHgt
    else ClientHeight:=DefHgt-paSunMoon.Height;
    lbxHolidays.Tag:=0;  // Update
    ShowKal (ShowDate);
    UpdateAutoStart;
    Invalidate;
    end;
  end;

procedure TfrmKalender.FormPaint(Sender: TObject);
begin
  JRCalendar1.Invalidate;
  JRCalendar2.Invalidate;
  JRCalendar3.Invalidate;
  JRCalendar4.Invalidate;
  JRCalendar5.Invalidate;
  JRCalendar6.Invalidate;
  ShowHolidays(AJahr);
  end;

procedure TfrmKalender.btnInfoClick(Sender: TObject);
begin
  with VersInfo do
    InfoDialog('',InternalName+' (Vers. '+Version+') - '+ProgVersDate+sLineBreak+
       CopyRight+sLineBreak+'E-Mail: '+EMailAdr+sLineBreak+
    'Die astronomischen Berechnungen basieren auf meiner Delphi-Umsetzung'+sLineBreak+
    'eines Javascripts von Arnold Barmettler (www.astronomie.info/www.CalSky.com)'+sLineBreak+
    'und den Algorithmen von Peter Duffett-Smith: "Practical Astronomy with your Calculator"');
  end;

procedure TfrmKalender.JRCalendarSelectCell(Sender: TObject; ACol,
  ARow: Integer; var CanSelect: Boolean);
var
  i,n  : integer;
begin
  with Sender as TJRCalendar do begin
    ShowDate:=CellDate[ACol,ARow];
    if MonthOf(ShowDate)<>Month then n:=-1
    else n:=HolidayCalendar.GetHolidayIndex(ShowDate,[hvHoliday,hvOther,hvPrivate]);
    if n<0 then lbxHolidays.ItemIndex:=-1
    else with lbxHolidays,Items do begin
      for i:=0 to Count-1 do
        if AnsiStartsText(HolidayCalendar.Holiday[n].Name,Strings[i]) then break;
      if i<Count then begin
        ItemIndex:=i; LastHolIndex:=ItemIndex;
        end;
      end;
    end;
  ShowSunMoon;
  end;

procedure TfrmKalender.ShowSunMoon;
var
  ss,sm : string;
  ssign : TStarSign;
  mphase : TLunarPhase;
begin
  if koSunMoon in Options then begin
    with HolidayCalendar.PosInfo do if length(LocName)>0 then ss:='Angaben fr '+LocName
    else ss:='Ort: '+DegToStr(Latitude,afDecMinutes,6,3)+'/'+DegToStr(Longitude,afDecMinutes,6,3);
    gbDay.Caption:=ss+' am '+DateToStr(ShowDate);
    GetSunMoonInfo(HolidayCalendar.PosInfo,ShowDate,ss,sm,ssign,mphase);
    edSun.Text:=ss; edMoon.Text:=sm;
    with imgSunSign do begin
      Picture:=nil;
      ilStarSigns.GetIcon(integer(ssign),Picture.Icon);
      Hint:=StarSigns[ssign];
      end;
    with imgMoonPhase do begin
      Picture:=nil;
      ilMoonPhases.GetIcon(integer(mphase),Picture.Icon);
      Hint:=MoonPhases[mphase];
      end;
    end;
  end;

procedure TfrmKalender.JRCalendarMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
var
  cpos    : TPoint;
  r       : TRect;
  ACol,
  ARow    : integer;
  dt      : TDateTime;
  s       : string;
begin
  with (Sender as TJRCalendar) do begin
    MouseToCell(X,Y,ACol,ARow);
    dt:=CellDate[ACol,ARow];
    if MonthOf(dt)<>Month then s:=''
    else if IsMarkDate then s:=MarkHint
    else s:=HolidayCalendar.GetHolidayNames(dt,htAll);
    if length(s)>0 then begin
      cpos:=Point(x,y);
      with r do begin
        TopLeft:= ClientToScreen(cpos);
        Top:=Top-20;
        Bottom:=Top+Canvas.TextHeight(s)+2;
        Right:=Left+Hintwin.Canvas.TextWidth(s)+10;
        end;
      HintWin.ActivateHint (r,s);
      end
    else HintWin.ReleaseHandle;
    end;
  end;

procedure TfrmKalender.FormMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
begin
  HintWin.ReleaseHandle;
  end;

procedure TfrmKalender.FormMouseWheelDown(Sender: TObject; Shift: TShiftState;
  MousePos: TPoint; var Handled: Boolean);
var
  n : integer;
begin
  with lbxHolidays do if BoundsRect.Contains(self.ScreenToClient(MousePos)) then begin
    n:=ClientHeight div ItemHeight;
    if TopIndex+n<Count-1 then TopIndex:=TopIndex+n;
    end
  else btnQPlusClick(Sender);
  Handled:=true;
  end;

procedure TfrmKalender.FormMouseWheelUp(Sender: TObject; Shift: TShiftState;
  MousePos: TPoint; var Handled: Boolean);
var
  n : integer;
begin
  with lbxHolidays do if BoundsRect.Contains(self.ScreenToClient(MousePos)) then begin
    n:=ClientHeight div ItemHeight;
    if TopIndex>=n then TopIndex:=TopIndex-n;
    end
  else btnQMinusClick(Sender);
  Handled:=true;
  end;

procedure TfrmKalender.UpdateHolidays;
var
  i,n : integer;
  dt : TDateTime;
begin
  dt:=EncodeDate(AJahr,AMonat,1);
  with lbxHolidays do begin
    n:=ClientHeight div ItemHeight;
    for i:=0 to Count-1 do if integer(Items.Objects[i])>dt then Break;
    if i<Count then TopIndex:=n*(i div n);
    end;
  end;

procedure TfrmKalender.lbxHolidaysMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  n : integer;
  j,m,w : word;
  s : string;
begin
  with lbxHolidays do begin
    n:=ItemAtPos(Point(x,y),true);
    if n>=0 then begin
      if LastHolIndex=n then begin
        ItemIndex:=-1;
        SetMarkDate(0);
        end
      else with Items do begin
        DecodeDate(integer(Objects[n]),j,m,w);
        if (m<AMonat) or (m>AMonat+5) then AMonat:=3*((m-1) div 3)+1;
        SetMarkDate(integer(Objects[n]));
        end;
      ShowKal (AMonat,AJahr);
      if ssCtrl in Shift then begin
        s:=Items[n]; s:=ReadNxtStr(s,':');
        if EditHolidayDialog.Execute (HolidayCalendar,s) then begin
          ItemIndex:=-1;
          SetMarkDate(0);
          lbxHolidays.Tag:=0;  // Update
          ShowKal (ShowDate);
          end;
        end;
      end
    else begin
      ItemIndex:=-1; SetMarkDate(0);
      ShowDate:=Date;
      ShowKal(ShowDate);
      end;
    LastHolIndex:=ItemIndex;
    end;
  end;

procedure TfrmKalender.btnTodayClick(Sender: TObject);
begin
  ShowDate:=Date;
  ShowKal(ShowDate);
  end;

procedure TfrmKalender.btnYearClick(Sender: TObject);
var
  n : integer;
begin
  n:=AJahr;
  if InputInteger (CursorPos,false,'Anderes Jahr','Jahr:','','',1,4,0,4000,false,n) then begin
    AJahr:=n; ShowKal (AMonat,AJahr);
    end;
  end;

procedure TfrmKalender.udJahrClick(Sender: TObject);
begin
  AJahr:=edJahr.Value;
  ShowKal (AMonat,AJahr);
  end;

procedure TfrmKalender.edJahrExit(Sender: TObject);
begin
  AJahr:=edJahr.Value;
  ShowKal (AMonat,AJahr);
  btnQPlus.SetFocus;
  end;

procedure TfrmKalender.edJahrKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if (Key=VK_TAB) or (Key=VK_RETURN) then begin
    AJahr:=edJahr.Value;
//    edJahr.Value:=AJahr;
    ShowKal (AMonat,AJahr);
    btnQPlus.SetFocus;
    end;
  end;

procedure TfrmKalender.edJahrKeyPress(Sender: TObject; var Key: Char);
begin
  if Key=#13 then Key:=#0;
  end;

procedure TfrmKalender.UpdateAutoStart;
var
  s : string;
begin
  (* Eintrag in Autostart anpassen *)
  s:=GetDesktopFolder(CSIDL_Startup);   (* Autostart-Ordner *)
  if length(s)>0 then begin
    s:=Erweiter(s,PrgName,'lnk');
    if FileExists(s) then DeleteFile(s);
    if (koAutoStart in Options) then begin
      MakeLink (s,Application.ExeName,'/'+sAuto,GetDesktopFolder(CSIDL_APPDATA),ProgVersName);
      end;
    end;
  end;

procedure TfrmKalender.WMTASKBAREVENT(var message: TMessage);
var
  point : TPoint;
begin
  case message.LParamLo of
    WM_LBUTTONDOWN   : begin
                       if not Visible then Show;
                       SetForegroundWindow(Handle); //Visible:=not Visible;
                       end;
    WM_RBUTTONDOWN   : if not Visible then begin
                         GetCursorPos(point);
                         Popupmenu.Popup(point.x,point.y);
                         end
                       else SetForegroundWindow(Handle);
      end;
  end;

procedure TfrmKalender.CloseItemClick(Sender: TObject);
begin
  Close;
  end;

procedure TfrmKalender.KalenderItemClick(Sender: TObject);
begin
  Show;
  end;

procedure TfrmKalender.btnTimeClick(Sender: TObject);
var
  z,d : integer;
begin
  GetTimeZoneInfo(z,d);
  TimeZoneDialog.Execute(z,d);
  end;

procedure TfrmKalender.btnCalcClick(Sender: TObject);
begin
  if DateCalcDialog.Execute(ShowDate,MarkHint) then SetMarkDate(ShowDate);
  end;

{ ------------------------------------------------------------------- }
procedure TfrmKalender.MsgHandler (var Msg : TMsg; var Handled : boolean);
begin
  if Msg.Message=WM_SHOW then begin
    if not visible then Show;
    Application.BringToFront;
    Handled:=true;
    end;
  end;

{ ------------------------------------------------------------------- }
{ ------------------------------------------------------------------- }
var
  mutex : THandle;
  h     : HWnd;

procedure TfrmKalender.btnPrintClick(Sender: TObject);
begin
  PrintPreviewDialog.Execute(AJahr,HolidayCalendar);
  end;

initialization
  Application.Title:=Prog;
  mutex := CreateMutex(nil,true,Prog);
  if getLastError = ERROR_ALREADY_EXISTS then begin
    h := 0;
    repeat
      h := FindWindowEx(0,h,'TApplication',PChar(Application.Title))
      until h <> application.handle;
    if h <> 0 then begin
      if IsWindowVisible(h) then begin
        (* Fenster in den Vordergrund *)
        Winapi.Windows.ShowWindow(h,SW_ShowNormal);
        Winapi.Windows.SetForegroundWindow(h);
        end
      else (* Fenster anzeigen *)
        Winapi.Windows.PostMessage (h,WM_SHOW,0,0);
      end;
    halt;
    end;

finalization
  ReleaseMutex(mutex);

end.
